#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import copy
import fwdpy11
import msprime
import numpy as np

import pytest
from fwdpy11_test_utilities import seed_list


class MutMonitor(object):
    def __init__(self, index, key):
        self.index = index
        self.key = key

    def __call__(self, pop, _):
        if pop.mcounts[self.index] == 0 or pop.mcounts[self.index] == 2 * pop.N:
            return True
        return False


def run_selective_sweep(
    msprime_seed,
    fp11_seed,
    ndescendants,
    N=500,
    alpha=1000,  # 2Ns
    rho=100,
    L=1.0,
    max_attempts=100,
):
    for initial_ts in msprime.sim_ancestry(
        samples=N,
        population_size=1000,
        recombination_rate=rho / 4 / N,
        random_seed=msprime_seed,
        sequence_length=1.0,
        num_replicates=max_attempts,
    ):
        pop = fwdpy11.DiploidPopulation.create_from_tskit(initial_ts)
        pdict = {
            "recregions": [fwdpy11.PoissonInterval(0, 1, 5e-2)],
            "gvalue": fwdpy11.Multiplicative(2.0),
            "rates": (0, 0, None),
            "prune_selected": False,
            "simlen": 10 * pop.N,
        }
        params = fwdpy11.ModelParams(**pdict)

        rng = fwdpy11.GSLrng(fp11_seed)
        data = fwdpy11.NewMutationData(effect_size=alpha / 2 / N, dominance=1.0)
        idx = pop.add_mutation(
            rng, ndescendants=ndescendants, data=data, window=(0.49, 0.51)
        )
        if idx is None:
            continue

        # Make sure we've chosen a valid time
        assert (
            pop.mutations[idx].g <= pop.tables.nodes[pop.tables.mutations[0].node].time
        ), f"{pop.mutations[idx].g} {pop.tables.nodes[pop.tables.mutations[0].node].time}"
        monitor = MutMonitor(idx, pop.mutations[idx].key)

        fixed = False
        pop_with_fixation = None

        while fixed is not True:
            pcopy = copy.deepcopy(pop)
            assert pcopy.generation == 0
            assert len(pcopy.mutations) == 1

            fwdpy11.evolvets(
                rng,
                pcopy,
                params,
                100,
                track_mutation_counts=True,
                stopping_criterion=monitor,
                suppress_table_indexing=True,
            )

            for f in pcopy.fixations:
                if f.key == monitor.key:
                    pop_with_fixation = pcopy
                    fixed = True

        return pop_with_fixation, idx
    return None, None


@pytest.mark.parametrize("msprime_seed", seed_list(135123, 5))
@pytest.mark.parametrize("fp11_seed", seed_list(5130125, 5))
def test_sweep_from_new_mutation(msprime_seed, fp11_seed):
    pop_with_fixation, idx = run_selective_sweep(msprime_seed, fp11_seed, 1)
    if pop_with_fixation is not None:
        assert pop_with_fixation.mcounts[idx] == 2 * pop_with_fixation.N
        _ = pop_with_fixation.dump_tables_to_tskit()


@pytest.mark.parametrize("msprime_seed", seed_list(135123, 5))
@pytest.mark.parametrize("fp11_seed", seed_list(5130125, 5))
@pytest.mark.parametrize("ndescendants", [2, 7, 10, 23, 100, 257])
def test_sweep_from_standing_variation(msprime_seed, fp11_seed, ndescendants):
    pop_with_fixation, idx = run_selective_sweep(msprime_seed, fp11_seed, ndescendants)
    if pop_with_fixation is not None:
        assert pop_with_fixation.mcounts[idx] == 2 * pop_with_fixation.N
        _ = pop_with_fixation.dump_tables_to_tskit()



#
# Copyright (C) 2021 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import json
import typing

import numpy as np
import tskit

from .._fwdpy11 import Mutation
from .metadata import (DiploidMetadata, decode_individual_metadata,
                       decode_mutation_metadata)


class WrappedTreeSequence(object):
    """
    Encapsulates a :class:`tskit.TreeSequence` as `self.ts`.

    Instance methods give additional functionality that usually
    requires interacting with the metadata.

    Instances of this class may be created using a :class:`tskit.TreeSequence`
    or a call to :func:`fwdpy11.tskit_tools.load`.

    .. versionadded:: 0.15.0
    """

    def _toplevel_metadata_value(self, name):
        if name in self._ts.metadata:
            return self._ts.metadata[name]
        return None

    def __init__(self, ts: tskit.TreeSequence):
        found = False
        for row in ts.provenances():
            record = json.loads(row.record)
            if "software" in record:
                if record["software"]["name"] == "fwdpy11":
                    found = True
        if not found:
            raise ValueError("this tree sequence was not generated by fwdpy11")

        self._ts = ts

    def timepoints_with_individuals(self, *, decode_metadata=False):
        """
        Return an iterator over all unique time points with individuals.

        For each time point a tuple of (time, nodes, metadata) is yielded.

        :param decode_individual_metadata: Whether to return decoded metadata.
        :type decode_individual_metadata: bool

        If `decode_individual_metadata` is `True`, metadata will be stored in
        a :class:`list` of :class:`fwdpy11.tskit_tools.DiploidMetadata`.
        If `False`, `None` will be yielded.
        """
        # Get rows of the node table where the nodes are in individuals
        nodes_in_individuals = np.where(self._ts.tables.nodes.individual != tskit.NULL)[
            0
        ]

        # Get the times
        node_times = self._ts.tables.nodes.time[nodes_in_individuals]

        unique_node_times = np.unique(node_times)

        for utime in unique_node_times[::-1]:
            # Get the node tables rows in individuals at this time
            x = np.where(node_times == utime)
            node_table_rows = nodes_in_individuals[x]
            assert np.all(self._ts.tables.nodes.time[node_table_rows] == utime)

            # Get the individuals
            individuals = np.unique(self._ts.tables.nodes.individual[node_table_rows])
            assert not np.any(individuals == tskit.NULL)

            if decode_metadata is True:
                # now, let's decode the individual metadata for this time slice
                decoded_individual_metadata = decode_individual_metadata(
                    self._ts.tables,
                    individuals,
                )
            else:
                decoded_individual_metadata = None
            yield utime, node_table_rows, decoded_individual_metadata

    def decode_individual_metadata(
        self, rows: typing.Optional[typing.Union[int, slice]] = None
    ) -> typing.List[DiploidMetadata]:
        """
        Decode individual metadata.

        See :func:`fwdpy11.tskit_tools.decode_individual_metadata` for details.
        """
        return decode_individual_metadata(self._ts.tables, rows)

    def decode_mutation_metadata(
        self, rows: typing.Optional[typing.Union[int, slice]] = None
    ) -> typing.List[typing.Optional[Mutation]]:
        """
        Decode mutation metadata.

        See :func:`fwdpy11.tskit_tools.decode_mutation_metadata` for details.
        """
        return decode_mutation_metadata(self._ts.tables, rows)

    @property
    def generation(self):
        return self._toplevel_metadata_value("generation")

    @property
    def seed(self):
        return self._toplevel_metadata_value("seed")

    @property
    def demes_graph(self):
        rv = self._toplevel_metadata_value("demes_graph")
        if rv is not None:
            import demes

            rv = demes.Graph.fromdict(rv)

        return rv

    @property
    def model_params(self):
        rv = self._toplevel_metadata_value("model_params")
        if rv is not None:
            import fwdpy11

            try:
                temp = {}
                for key, value in rv.items():
                    temp[key] = fwdpy11.ModelParams(**eval(value))
                rv = temp
            except:
                rv = fwdpy11.ModelParams(**eval(rv))

        return rv

    @property
    def data(self):
        rv = self._toplevel_metadata_value("data")
        if rv is None:
            return rv

        return rv

    @property
    def ts(self):
        return self._ts
